import os

from setuptools import setup

setup(
    name='jarbas_hive_mind',
    version='0.3.5',
    packages=['jarbas_hive_mind',
              'jarbas_hive_mind.minds',
              'jarbas_hive_mind.minds.simple',
              'jarbas_hive_mind.nodes',
              'jarbas_hive_mind.nodes.flask',
              'jarbas_hive_mind.utils',
              'jarbas_hive_mind.utils.messagebus',
              'jarbas_hive_mind.drones',
              'jarbas_hive_mind.drones.simple',
              'jarbas_hive_mind.bridges',
              'jarbas_hive_mind.bridges.hackchat',
              'jarbas_hive_mind.bridges.facebook',
              'jarbas_hive_mind.bridges.twitch',
              'jarbas_hive_mind.bridges.twitter',
              'jarbas_hive_mind.database',
              'jarbas_hive_mind.terminals',
              'jarbas_hive_mind.terminals.https',
              'jarbas_hive_mind.terminals.cli',
              'jarbas_hive_mind.terminals.basic_web',
              'jarbas_hive_mind.terminals.speech',
              'jarbas_hive_mind.terminals.speech.stt',
              'jarbas_hive_mind.terminals.speech.recognizer',
              'jarbas_hive_mind.terminals.webchat'],
    include_package_data=True,
    install_requires=["pyopenssl",
                      "service_identity",
                      "flask",
                      "flask-sslify",
                      "yagmail",
                      "autobahn",
                      "twisted",
                      "sqlalchemy",
                      "lingua_franca",
                      "requests",
                      "python-dateutil==2.6.0",
                      "tornado==4.5.3",
                      "pyee==5.0.0",
                      "websocket-client==0.54.0"],
    url='https://github.com/JarbasAl/hive_mind',
    license='MIT',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='Networking utilities for mycroft core'
)
